import pathlib
import sys

import pytest


@pytest.fixture
def collector():
    return []


@pytest.fixture
def sample_project(tmpdir):
    """Return a sample project."""
    tmpdir = pathlib.Path(tmpdir)
    (tmpdir / "sample_project").mkdir()
    (tmpdir / "sample_project" / "__init__.py").write_text("")
    (tmpdir / "sample_project" / "services.py").write_text(services)
    (tmpdir / "sample_project" / "users").mkdir()
    (tmpdir / "sample_project" / "users" / "__init__.py").write_text("")
    (tmpdir / "sample_project" / "users" / "controllers.py").write_text(controllers)
    (tmpdir / "sample_project" / "users" / "cli.py").write_text(cli)
    sys.path.insert(0, tmpdir.as_posix())
    yield tmpdir
    sys.path = sys.path[1:]


services = """
class App():
    app = None
    def init_app(self, app):
        self.app = app

foo = App()
bar = App()
"""

controllers = """
from flask import Blueprint

blueprint = Blueprint("users", __name__)

@blueprint.route("/", methods=["GET"])
def users():
    return ''
"""

cli = """
from flask.cli import AppGroup

app_group = AppGroup("users")
"""



__author__ = """Roman Imankulov"""
__email__ = "roman.imankulov@gmail.com"
__version__ = "1.1.0"


from deescovery.discovery import IRule, ModuleRule, ObjectRule, discover

__all__ = (
    "discover",
    "IRule",
    "ModuleRule",
    "ObjectRule",
)



import importlib
import logging

logger = logging.getLogger(__name__)


def import_module(module_name: str) -> None:
    """Import module by its name.

    Calls importlib.import_module() and logs the action with logging.debug().
    """
    logging.debug("Importing module %s", module_name)
    importlib.import_module(module_name)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['deescovery', 'tests']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'deescovery',
    'version': '1.1.0',
    'description': 'Discover packages and classes in a python project.',
    'long_description': '<!--intro-start-->\n# Deescovery\n\n**Deescovery** is a Python package to find and initialize modules of your Python projects on startup.\n\n- Find and register blueprints in a Flask project.\n- Automatically initialize Flask extensions.\n- Find all SQLAlchemy models to make alembic happy.\n- Find all FastAPI endpoints.\n- Collect all Celery tasks.\n\nInitially designed to initialize Flask applications, it was made generic enough to work with any micro-framework or no framework at all.\n\n## Micro-framework initialization problem\n\nMicro-framework-based projects are clean while they\'re small. Every micro-framework codebase I\'ve seen, has a mess in the project initialization. With time, `create_app()` becomes filled with ad-hoc settings, imports-within-functions, and plug-in initializations.\n\nThe Application Factory Pattern, proposed, for example, in the [official Flask documentation](https://flask.palletsprojects.com/en/2.0.x/patterns/appfactories/), and the [Flask Mega-Tutorial](https://blog.miguelgrinberg.com/post/the-flask-mega-tutorial-part-xv-a-better-application-structure), legitimize this approach.\n\nThe nature of `create_app()` leaves no place for the [open-closed principle](https://blog.cleancoder.com/uncle-bob/2014/05/12/TheOpenClosedPrinciple.html). We update this module every time we add a new plug-in, a new blueprint, or a new package.\n\n```python\n# myproject/__init__.py\n\ndef create_app(config_class=Config):\n    app = Flask(__name__)\n    app.config.from_object(config_class)\n\n    db.init_app(app)\n    migrate.init_app(app, db)\n    login.init_app(app)\n    mail.init_app(app)\n    bootstrap.init_app(app)\n    moment.init_app(app)\n    babel.init_app(app)\n\n    from myproject.errors import bp as errors_bp\n    app.register_blueprint(errors_bp)\n\n    from myproject.auth import bp as auth_bp\n    app.register_blueprint(auth_bp, url_prefix=\'/auth\')\n\n    return app\n```\n\n_A common Flask application. The code is based on the Flask Mega-Tutorial._\n\nWith `deescovery`, you can make the same code shorter, and remove the dependencies from implementation details.\n\n```python\n# file: myproject/app.py\nfrom flask import Flask\nfrom deescovery import discover\nfrom deescovery.flask import get_flask_rules\n\n\ndef create_app():\n    flask_app = Flask(__name__)\n    flask_app.config.from_object("myproject.config")\n    discover("myproject", get_flask_rules("myproject", flask_app))\n    return flask_app\n```\n\n\n<!--intro-end-->\n\n## Read more\n\n- [Usage with Flask](https://imankulov.github.io/deescovery/flask/)\n- [Usage with anything else](https://imankulov.github.io/deescovery/anything_else/)\n- [API](https://imankulov.github.io/deescovery/api/deescovery/)\n',
    'author': 'Roman Imankulov',
    'author_email': 'roman.imankulov@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/imankulov/deescovery',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7.8,<4.0',
}


setup(**setup_kwargs)



"""Module for the Flask-specific rules.

A sample Flask app generator would look like this.

```python
from flask import Flask
from deescovery discover
from deescovery.flask import get_flask_rules

def app() -> Flask:
    flask_app = Flask(__name__)
    flask_app.config.from_object("myapp.config")
    discovery_rules = get_flask_rules("myapp", flask_app)
    discover("dashboards", discovery_rules)
    return flask_app
```
"""
from typing import List

from deescovery import IRule
from deescovery.discovery import ModuleRule, ObjectRule
from deescovery.matchers import MatchByMethod, MatchByPattern, MatchByType


def get_flask_rules(import_path: str, flask_app) -> List[IRule]:
    """Return a list of rules useful for the Flask application.

    The following rules will be returned:

    - Load SQLAlchemy models (files models.py)
    - Load Flask blueprints (files controllers.py)
    - Load Flask CLI commands (files cli.py)
    - Initialize services (top-level file services.py)

    The service initialization step relied on the de-facto standard for
    initializing Flask extensions: calls `obj.init_app(flask_app)`
    for each of the objects in the top-level `services.py` file.

    Args:
        import_path: name of the top-level module of the project (like, "myproject")
        flask_app: a Flask app instance.

    Returns:
        A list of rules, suitable to be passed to "deescovery.discover()"
    """
    return [
        models_loader(import_path),
        blueprints_loader(import_path, flask_app),
        commands_loader(import_path, flask_app),
        service_initializer(import_path, flask_app),
    ]


def models_loader(import_path):
    """Load all models."""
    return ModuleRule(
        name="Flask models loader",
        module_matches=MatchByPattern(generate_patterns(import_path, "models")),
    )


def blueprints_loader(import_path, flask_app):
    """Find and import all blueprints in the application."""
    try:
        from flask import Blueprint
    except ImportError:
        raise RuntimeError("Flask is not installed.")
    return ObjectRule(
        name="Flask blueprints loader",
        module_matches=MatchByPattern(generate_patterns(import_path, "controllers")),
        object_matches=MatchByType(Blueprint),
        object_action=flask_app.register_blueprint,
    )


def commands_loader(import_path, flask_app):
    """Find all commands and register them as Flask CLI commands."""
    try:
        from flask.cli import AppGroup
    except ImportError:
        raise RuntimeError("Flask is not installed.")
    return ObjectRule(
        name="Flask CLI commands loader",
        module_matches=MatchByPattern(generate_patterns(import_path, "cli")),
        object_matches=MatchByType(AppGroup),
        object_action=flask_app.cli.add_command,
    )


def service_initializer(import_path, flask_app):
    """Find and initialize all instances of Flask applications.

    Notice that the initialize scans for top-level services files, and doesn't
    walk over all your app's domain package.
    """
    return ObjectRule(
        name="Flask service initializer",
        module_matches=MatchByPattern([f"{import_path}.services"]),
        object_matches=MatchByMethod("init_app"),
        object_action=lambda obj: obj.init_app(app=flask_app),
    )


def generate_patterns(import_path: str, module_prefix: str) -> List[str]:
    """Generate a list of patterns to discover.

    For example, gen_patterns("myapp", "models") generates patterns that make matchers
    discover the content in the following files.

        myapp/users/models.py
        myapp/invoices/models.py
        (etc. for all domain packages beyond "users" and "invoices")
        ...

        myapp/users/models_roles.py
        myapp/users/models_groups.py
        (etc. for all modules started with "models_" in all domain packages)
        ...

        myapp/users/models/roles.py
        myapp/users/models/groups.py
        (if you prefer nested structures)
    """
    return [
        f"{import_path}.*.{module_prefix}",
        f"{import_path}.*.{module_prefix}_*",
        f"{import_path}.*.{module_prefix}.*",
    ]
